# This file is part of the xxdiff package.  See xxdiff for license and details.

"""xx-encrypted [<options>] <encrypted-file> [<encrypted file> ...]

Compare and merge contents of encrypted files relatively safely.

This script wraps around xxdiff, first decrypting the input files to temporary
files (for a short time) and running xxdiff on these files.  There are two
typical uses of this program:

1) it is used to compare two encrypted files.  With the --output option, a
   decision is required and an encrypted version of the merged file is output to
   the specified file and the merged file deleted promptly.  Note that without
   the --output option, even if the merged file is saved, it is deleted once
   xxdiff exits.

2) it is used to split and resolve CVS conflicts in an armored encrypted file
   (see --unmerge option).  The merged file is encrypted and output over the
   conflictual input file (i.e. it replaces it with the encrypted version of the
   merged file).  This is very useful if you maintain armored encrypted files in
   CVS repositories because otherwise an encrypted file with a CVS conflict in
   it becomes useless.

Using gpg-agent
---------------

Usage of this program with password caching using gpg-agent makes it much easier
to call on multiple files.  The user's password given key is asked only once by
gpg-agent, kept in memory, and then decryption occurs without user intervention.

Merging Encrypted Files with Mercurial
--------------------------------------

Mercurial can be easily configured so that this tool gets invoked automatically
when encrypted files need to be merged. This way you can independently
edit encrypted files using Emacs (I use a custom mailcrypt setup to seamlessly
decode on open and encode on save) and when tracked in a Mercurial repository,
if the encrypted files have been edited in two separate revisions, xx-encrypted
is invoked automatically and brings up a 3-way xxdiff to merge them interactively.
Here's the configuration I use in my .hgrc::

  [merge-tools]
  xx-encrypted =
  xx-encrypted.priority = 100
  xx-encrypted.premerge = False
  xx-encrypted.args = $local $base $other -o $output

  [merge-patterns]
  **.asc = xx-encrypted
  **.gpg = xx-encrypted

Alternatively, you may just use xx-hg-merge which deals with all file types,
encrypted or not:

  [merge-patterns]
  ** = xx-hg-merge

"""



__moredoc__ = """
Safety Notes
------------

The encrypted files are decrypted to temporary files for a short amount of time,
and are deleted when xxdiff appears.  Note that their deletion is as safe as
Python's tempfile module allows it to be (in the author's opinion, safe
enough). I left comments in the code to allow a user to review where the files
are decrypted so they can judge by themselves if it is safe enough for their
use.

We could do much better in terms of safety if we could feed the input files to
xxdiff through different file descriptors (not impossible to implement) AND
calculate the diffs internally.

(Note that if someone can manipulate which program is used to actually perform
the diffs (e.g. modifying an unsuspecting user's resources in ~/.xxdiffrc), they
could feed the decrypted files to an arbitrary program.)
"""

__author__ = "Martin Blais <blais@furius.ca>"
__depends__ = ['xxdiff', 'Python-2.4', 'GnuPG']

# stdlib imports.
import sys, os, re
from os.path import *
from tempfile import NamedTemporaryFile
from subprocess import Popen, PIPE

# xxdiff imports.
import xxdiff.scripts
import xxdiff.invoke
import xxdiff.scm.cvs
from xxdiff.scripts import tmpprefix


decodecmd = '%(gpg)s --decrypt --use-agent '
encodecmd_noarmor = '%(gpg)s --encrypt --use-agent '
encodecmd = encodecmd_noarmor + '--armor '


def get_recipient(text, gpg):
    """ Extract the recipient/key name from the given encrypted text, without
    saving any temporary file. """
    p = Popen(('gpg', '--list-only'), stdin=PIPE, stdout=PIPE, stderr=PIPE, text=True)
    out, err = p.communicate(text)
    mo = re.compile('.*public key is ([0-9A-F]+)', re.M).search(err)
    if mo:
        return mo.group(1)


def diff_encrypted(textlist, opts, outmerged=None):
    """
    Run a comparison of the encrypted texts specified in textlists and if an
    'outmerged' filename is specified, encrypt the merged file into it. Note
    that the texts are not filenames, but actual contents of files.
    """
    m = {'gpg': opts.gpg}

    # Create temporary files.
    tempfiles = []
    for t in range(len(textlist)):
        f = NamedTemporaryFile(prefix=tmpprefix)
        print('== TEMPFILE', f.name)
        tempfiles.append(f)

    # Figure out the key/recipient for the first file.
    if opts.recipient is None:
        opts.recipient = get_recipient(textlist[0], opts.gpg)

    # Always create a temporary file for the merged file, we will delete it for
    # sure, since it would contain decrypted content if saved.

    # Decode the files.
    for i in range(len(textlist)):
        t, f = textlist[i], tempfiles[i]

        # Decode one file to an existing temporary file.
        fin, fout = os.popen2(decodecmd % m, 'w')
        fin.write(t)
        fin.close()
        decoded_output = fout.read()
        fout.close()

        f.write(decoded_output)
        f.flush()

    # Spawn xxdiff on the temporary/decoded files.
    waiter = xxdiff.invoke.xxdiff_decision(
        opts, nowait=1, *[x.name for x in tempfiles])

    # Close and automatically delete the temporary/decoded files.
    for f in tempfiles:
        f.close()

    print('Waiting...')
    decision, mergedf, retcode = waiter()

    if decision != 'NODECISION' and outmerged:
        # Read the decoded merged output file from xxdiff.
        textm = mergedf.read()
        assert textm

        # Close and automatically delete the decoded merged output file.
        mergedf.close()

        # Encode the merged output text.
        if not opts.dont_armor:
            cmd = encodecmd % m
        else:
            cmd = encodecmd_noarmor % m

        if opts.recipient:
            cmd += ' --recipient "%s"' % opts.recipient

        fin, fout = os.popen2(cmd, 'w')
        fin.write(textm)
        fin.close()
        encoded_output = fout.read()
        fout.close()

        # Write out the encoded output file.
        try:
            f = open(outmerged, 'w')
            f.write(encoded_output)
            f.close()
        except IOError as e:
            print('Error: cannot write to encoded merged file.', file=sys.stderr)
            raise e

    return decision


def parse_options():
    """
    Parse the options.
    """
    import optparse
    parser = optparse.OptionParser(__doc__)

    xxdiff.invoke.options_graft(parser)

    parser.add_option('-g', '--gpg', default="gpg",
                      help="Specify path to gpg program to use.")

    parser.add_option('-A', '--dont-armor', action='store_true',
                      help="Create output file in binary format.")

    parser.add_option('-o', '--output', action='store',
                      help="Require and encrypt merged output.")

    parser.add_option('-u', '--unmerge', action='store_true',
                      help="Split CVS conflicts in single input file and "
                      "encrypt required output merged file over input.")

    parser.add_option('-r', '--recipient', action='store',
                      help="Encrypt for user id name.")

    global opts
    xxdiff.scripts.install_autocomplete(parser)

    opts, args = parser.parse_args()

    if not args:
        raise parser.error('No files to decrypt and compare.')

    xxdiff.invoke.options_validate(opts, parser, logs=sys.stdout)

    return opts, args

def encrypted_main():
    opts, args = parse_options()

    if isabs(opts.gpg) and not exists(opts.gpg):
        raise SystemExit('Error: gpg program does not exist in "%s"' %
                         opts.gpg)

    if not (opts.unmerge or opts.output):
        if opts.dont_armor:
            print("Warning: there will be no output file, " +
                  "--dont-armor will means nothing special.", file=sys.stderr)
        if opts.recipient:
            print("Warning: there will be no output file, " +
                  "--recipient will means nothing special.", file=sys.stderr)

    if opts.unmerge:
        for fn in args: # do all files specified on cmdline, why not.
            # Read input conflict file.
            text = open(fn, 'r').read()
            text1, text2 = xxdiff.scm.cvs.unmerge2(text)
            diff_encrypted([text1, text2], opts, fn)
    else:
        if len(args) <= 1:
            raise SystemExit("Error: you need to specify 2 or 3 arguments.")

        textlist = []
        for fn in args:
            text = open(fn, 'r').read()
            textlist.append(text)

        diff_encrypted(textlist, opts, opts.output)


main = encrypted_main

if __name__ == '__main__':
    main()



# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
Functions for Rational ClearCase.
"""

__author__ = 'Martin Blais <blais@furius.ca>'


# stdlib imports.
from subprocess import Popen, call, PIPE

# xxdiff imports.
from xxdiff.scripts import script_name


def is_checked_out(filename):
    """
    Check if the file is already checked out.
    """
    p = Popen(['cleartool', 'desc', '-s', filename], stdout=PIPE, stderr=PIPE, text=True)
    stdout, stderr = p.communicate()
    if p.returncode != 0:
        raise SystemExit( stderr )
    
    if stdout.strip().endswith('CHECKEDOUT'):
        return True
    
    return False
    
def checkout(filename):
    """
    Checkout the given filename from ClearCase.
    """
    msg = 'Checked out by %s' % script_name

    call(['cleartool', 'co', '-comment', msg, filename])
    # FIXME: TODO -- you need to check the return code.


def commit(filename):
    """
    Commit the given filename into CVS.
    """
    msg = 'Committed by %s' % script_name
    call(['cleartool', 'commit', '-command', msg, filename])





# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
Functions for CVS.
"""

__author__ = 'Martin Blais <blais@furius.ca>'


# stdlib imports.
import os, re
from subprocess import call


def is_checked_out(filename):
    """
    Return True if the file is checked out.
    """
    return os.access(os.W_OK)


def checkout(filename):
    """
    Checkout the given filename from CVS.
    """
    call(['cvs', 'edit', filename])


def commit(filename):
    """
    Commit the given filename into CVS.
    """
    call(['cvs', 'commit', filename])


# Note: this is cut-n-paste from cvs.py, for simplicity of distribution.
def unmerge2(text):
    """
    Unmerges conflicts between two files and returns the two split original
    files.  This is not meant to work with three files.
    """
    begre = re.compile('^<<<<<<< (.*)')
    sepre = re.compile('^=======\s*$')
    endre = re.compile('^>>>>>>> (.*)')

    OUT, IN1, IN2 = 0, 1, 2

    status = OUT
    lines1, lines2 = [], []
    for l in text.splitlines():
        if status == OUT:
            if begre.match(l):
                status = IN1
            else:
                lines1.append(l)
                lines2.append(l)
        elif status == IN1:
            if sepre.match(l):
                status = IN2
            else:
                lines1.append(l)
        elif status == IN2:
            if endre.match(l):
                status = OUT
            else:
                lines2.append(l)

    text1 = os.linesep.join(lines1)
    text2 = os.linesep.join(lines2)

    return text1, text2




# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
Functions to select files within directories.
"""



__author__ = 'Martin Blais <blais@furius.ca>'


# stdlib imports.
import sys, os, re, optparse
from os.path import join, isfile, exists


def options_graft(parser):
    """
    Adds a new option group for file selection options to the given parser and
    returns this new group.
    """
    group = optparse.OptionGroup(parser, "File selection options",
                                 "These options affect which files are "
                                 "selected for grepping in the first place.")

    def regexp_cb(option, opt_str, value, parser):
        "Compile and set the regular expression"
        try:
            regexp = re.compile(value)
            getattr(parser.values, option.dest).append(regexp)
        except re.error as e:
            raise optparse.OptionValueError(
                "compiling regular expression '%s': %s" % (value, e))

    for oshort, olong, ohelp in (
        ('-s', '--select',
         "Adds a regular expression for filenames to process."),
        ('-I', '--ignore',
         "Adds a regular expression for filenames to ignore."),
        ):
        group.add_option(oshort, olong,
                         action='callback', type='str', callback=regexp_cb,
                         metavar="REGEXP", default=[],
                         help=ohelp)

    # Note: we could improve this as soon as Python will include a more recent
    # version of Optik, which implements the action 'append_const'.
    def append_const_cb(const):
        "Create callback to implement an append_const action via a callback."
        def cb(option, opt_str, value, parser):
            getattr(parser.values, option.dest).append(const)
        return cb

    for fname, ftype, fregexp in (('C++', 'cpp', '.*\.(h(pp)?|c(pp|\+\+|c)?)$'),
                                  ('Python', 'py', '.*\.py$')):

        group.add_option('--%s' % ftype, '--select-%s' % ftype, dest='select',
            action='callback', callback=append_const_cb(re.compile(fregexp)),
            default=[],
            help=("Adds a regular expression for selecting %s files to "
                  "match against.") % fname)

    group.add_option('--select-grep', action='store',
                     metavar='REGEXP',
                     help="Further restrict the files to those which match "
                     "the given regular expression.")

    group.add_option('--ignore-grep', action='store',
                     metavar='REGEXP',
                     help="Do not select files to those which match "
                     "the given regular expression.")

    group.add_option('-f', '--select-from-file', action='store',
                     metavar='FILE',
                     help="Do not recurse through directories to find files "
                     "but instead read the list of filenames from the "
                     "given file.")

    group.add_option('-r', '--root', action='append', dest='roots',
                     default=[], metavar='ROOT',
                     help="Specify a root to perform the search from "
                     "(default is CWD).  You can use this option many times"
                     "for multiple roots.")

    # This is a convenient option rather than making a test script, because we
    # don't pollute the bin directory with a script that does not do much, and
    # the option is available in all the scripts that use the select files
    # proces.
    group.add_option('-T', '-@', '--select-debug', action='store_true',
                     help="Only list the files and exit.  This is used to "
                     "debug and test out which files will match your "
                     "replacement process when you run it.")

    group.add_option('-N', '--select-no-defaults', action='store_true',
                     help="Do not setup default ignores (i.e. by default, "
                     "appropriate ignore patterns are set for Subversion "
                     "and CVS directories, etc.)")

    parser.add_option_group(group)

    return group


ignore_defaults = ('CVS', '\.(svn|git|hg|bzr)$')

def options_validate(opts, parser, logs=None):
    """
    Validate and prepare the parsed options for the file selection option group.
    This method returns an appropriate generator for selecting files.
    """
    # Process file selection options
    if opts.select_from_file:
        if opts.select or opts.ignore:
            parser.error("You cannot use select-from-file and other "
                         "select options together.")
        if opts.roots:
            parser.error("You cannot use select-from-file and specify "
                         "a list of root directories to search for.")
        # Note: eventually we might want to support chaining the generators
        # instead.
    else:
        # If no root directory has been specified, use the CWD.
        if not opts.roots:
            opts.roots.append(os.getcwd())
        else:
            missing = [x for x in opts.roots if not exists(x)]
            if missing:
                parser.error("Some root directories do not exist: %s" %
                             ' '.join(missing))

        # By default ignore some common directories.
        if not opts.select_no_defaults:
            opts.ignore = list(map(re.compile, ignore_defaults)) + opts.ignore

        # Process all files if no filter specified
        if not opts.select:
            opts.select.append(re.compile('.*'))

    # Compile the regular expressions if given.
    for regname in 'select_grep', 'ignore_grep':
        regstr = getattr(opts, regname)
        if regstr:
            try:
                setattr(opts, regname, re.compile(regstr))
            except re.error as e:
                parser.error("Error compiling regexp '%s':\n%s" % (regstr, e))

    # Create an appropriate generator for the "select" method
    if opts.select_from_file:
        selector = select_from_file(opts.select_from_file)
    else:
        selector = select_patterns(opts.roots, opts)

    # If we're asked to print the list of files, exhaust the generator and print
    # the filenames and exit.
    if opts.select_debug:
        for fn in selector:
            print(fn)
        sys.exit(0)

    return selector


def select_patterns(rootdirs, opts):
    """
    Generator that selects files to process by regexps.

    Arguments:
    - rootdirs: root directories -> list of strings
    - select: patterns to select -> list of re regexp objects
    - ignore: patterns to ignore -> list of re regexp objects

    Yields: selected filenames -> string
    """
    # Walk the tree of files and select files as requested
    for root in rootdirs:
        for dn, dirs, files in os.walk(root):
            # Filter directories.
            for d in list(dirs):
                add = True
                # Only ignore applies to directories.
                for ire in opts.ignore:
                    if ire.match(d):
                        add = False
                        break
                if not add:
                    dirs.remove(d)

            # Filter files.
            for fn in files:
                # Skip symlinks and special files.
                if not isfile(join(dn, fn)):
                    continue

                add = False

                # Select.
                for sre in opts.select:
                    if sre.match(fn):
                        add = True
                        break

                # Further restrict by grepping the file for a pattern.
                if add and opts.select_grep:
                    try:
                        text = open(join(dn, fn), 'r').read()
                    except IOError as e:
                        raise SystemExit(
                            "Error: could not read file '%s' for grep." % fn)
                    if not opts.select_grep.search(text):
                        add = False

                # Ignore files.
                for ire in opts.ignore:
                    if ire.match(fn):
                        add = False
                        break

                # Further ignore by grepping the file for a pattern.
                if add and opts.ignore_grep:
                    try:
                        text = open(join(dn, fn), 'r').read()
                    except IOError as e:
                        raise SystemExit(
                            "Error: could not read file for ignore '%s'." % fn)
                    if opts.ignore_grep.search(text):
                        add = False

                if add:
                    yield join(dn, fn)

def select_from_file(fn):
    """
    Generator that yields selected filenames read from a file.

    Arguments:
    - fn: filename of the file that contains the list of filenames -> string

    Yields: selected filenames -> string
    """
    try:
        for fn in open(fn, 'r'):
            yield fn.strip()
    except IOError as e:
        raise SystemExit("Error: cannot open/read file (%s)" % e)


def test():
    """
    Test stub for selection options.
    """
    parser = optparse.OptionParser(__doc__.strip())
    options_graft(parser)
    opts, args = parser.parse_args()

    selector = options_validate(opts, parser)

    print('from-file:', opts.select_from_file)
    print('select:', [x.pattern for x in opts.select])
    print('ignore:', [x.pattern for x in opts.ignore])
    print('roots:', args)
    print()

    for fn in selector:
        print(fn)


if __name__ == '__main__':
    test()



# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
Misc utilities.  No, really, they are really miscallenous, and unrelated.
"""

__author__ = 'Martin Blais <blais@furius.ca>'


# stdlib imports.
import os, re
from os.path import exists
from curses.ascii import isalnum
from subprocess import Popen, PIPE


def consepairs(seq):
    """
    Iterate over consecutive pairs.
    """
    it = iter(seq)
    prev = next(it)
    for el in it:
        yield prev, el
        prev = el


def idify(s, strip=True, preserve_chars=[]):
    """
    Removes non-alphanumeric characters from a string, morphing them into
    underscores.
    """
    ss = ''
    preserve_list = ['\n'] + preserve_chars
    for c in s:
        if isalnum(c):
            ss += c
        elif c in preserve_list:
            ss += c
        else:
            ss += '_'
    if strip:
        b, e = 0, len(ss)
        while b < len(ss) and ss[b] == '_': b += 1
        while e > 1 and ss[e-1] == '_': e -= 1
        ss = ss[b:e]
    return ss


# Note: there has been a 'file' command for a long time under UNIX.  We favor
# the short options to promote portability.  The --bried and --dereference
# options were taken from Ian F. Darwin's file implementation.
guesscmd = ['file', '-b', '-L']
text_re = re.compile('\\btext\\b')
xml_re = re.compile('\\bXML\\b')
empty_re = re.compile('^empty$')

def istextfile(fn):
    """
    Attempts to guess if the file indicated by the given filename is a text file
    or a binary file.
    """

    # Unfortunately 'file' does not return an appropriate return code when there
    # is an error.  On top of that, it returns its errors on stdout.  We need to
    # parse the output.
    p = Popen(guesscmd + [fn], stdout=PIPE, stderr=PIPE, text=True)
    stdout, stderr = p.communicate()
    if p.returncode != 0 or stderr or stdout.startswith('cannot open'):
        raise RuntimeError("Error: running 'file' on '%s'." % fn)

    return bool(text_re.search(stdout) or xml_re.search(stdout) or empty_re.match(stdout))


def makedirs(dirn, error_on_exist=True):
    """
    A replacement for os.makedirs() that has an option to avoid complaining if
    the destination directory already exists.
    """
    if error_on_exist is False:
        if exists(dirn):
            return 
    os.makedirs(dirn)




# This file is part of the xxdiff package.  See xxdiff for license and details.

"""xx-patch [<options>] [<patchfile>]

This simple script splits the input patch for the individual files it
represents, applies the individual patches to temporary files and for each,
spawns an xxdiff for it.  This effectively allows you merge the patch
interactively.  You can accept or reject a patch for each file.  You can also
merge the patch interactively with xxdiff, and select for the merged results to
be applied on the output file.

If no <patchfile> is specified, xx-patch attempts to read the patch from
stdin.

Important: the new (patched) file will always be shown on the right.  The
patched file will display (patched) in its filename title.  When a decision is
asked to the user, ACCEPT always means keep/save the file that appears on the
right.  The script takes the following actions upon the answer:

- ACCEPT: keep/copy the file on the right side to the output file;
- REJECT: keep/copy the file on the left side to the output file;
- MERGED: copy the merged file on the output file;

If a patch for a file is accepted or merged, before being overwritten, a copy of
the original files are saved with the extension .orig, like patch itself.

Note: notice that the --reverse and --invert options below are implementing
orthogonal behaviour.  It can be confusing to understand the differences
between these two.  If the a reverse patch is indeed reverting changes (and
not constructed in the incorrect order), applying the patch will revert the
changes as well (the display order is not changed).  This behaviour is
exactly like that of patch.  Using invert, however, changes the logic of
what this script does upon accepting or rejecting a patch.  In any case, you
will always find the 'new file if accepted' on the right side and the
patched file is indicated with a string in the filename title.

Examples
--------

  To apply a patch, simply cd into the root of the directory where the patch
  should be applied, and invoke like this:

     cat patchfile | xx-patch

  To preview the changes that have been made in a cvs checkout area, use the
  following command::

     cvs diff | xx-patch --reverse --invert --preview

  (For better cvs-specific behaviour about committing changes, see script
  xx-cvs-diff.)

.. important::

   This script is unfinished, we still need to implement file addition/removal
   support.

"""



__author__ = "Martin Blais <blais@furius.ca>"
__depends__ = ['xxdiff', 'Python-2.4', 'diffutils']

# stdlib imports.
import sys, os
import shutil
from tempfile import NamedTemporaryFile

# xxdiff imports
from xxdiff.scripts import tmpprefix
import xxdiff.patches
import xxdiff.invoke
import xxdiff.backup


def parse_options():
    """
    Parse the options.
    """
    import optparse
    parser = optparse.OptionParser(__doc__.strip())

    xxmodules = xxdiff.backup, xxdiff.invoke
    for mod in xxmodules:
        mod.options_graft(parser)

    parser.add_option('-p', '--patch-options', action='store', metavar="OPTS",
                      default='',
                      help="specifies additional options to pass on to patch.")

    parser.add_option('-s', '--strip', action='store', type='int',
                      help="strip option to patch, provided for convenience.")

    parser.add_option('-R', '--reverse', action='store_true',
                      help="""apply the patch as a reverse patch (see
                      patch(1) option --reverse).  The patched file will be
                      shown as the new file regardless.""")

    parser.add_option('-I', '--invert', action='store_true',
                      help="""change the order of the displayed files, so that
                      the patched file shows as the old file.  Accepting a
                      change will thus do nothing and rejecting a change
                      will copy the patched file over.""")

    parser.add_option('-n', '--dry-run', '--preview', action='store_true',
                      help="""don't apply the patch, just show the
                      differences (you don't have to make any decisions).""")

    xxdiff.scripts.install_autocomplete(parser)
    opts, args = parser.parse_args()

    for mod in xxmodules:
        mod.options_validate(opts, parser, logs=sys.stdout)

    return opts, args


def patch_main():
    """
    Main program for patch script.
    """
    opts, args = parse_options()

    #
    # read the input file.
    #
    if not args:
        args = ['-']

    chunks = []
    for fn in args:
        try:
            if fn == '-':
                iif = sys.stdin
            else:
                iif = open(fn, 'r')
            text = iif.read()
            iif.close()
        except IOError as e:
            raise SystemExit(
                "Error: reading patchfile %s: %s" % (fn, str(e)))

        chunks += xxdiff.patches.splitpatch(text)

    #
    # For each subpatch, apply it individually
    #
    for filename, patch in chunks:
        # print patch contents for this file.
        print()
        print()
        print("FILE:", filename)
        print('*' * 40)
        print(patch)
        print('*' * 40)

        # feed diffs to patch, patch will do its deed and save the output to
        # a temporary file.
        tmpfp = NamedTemporaryFile(prefix=tmpprefix)

        popts = opts.patch_options
        popts += ['--reverse', ''][ opts.reverse == None ]
        if opts.strip:
            popts += '--strip=%d' % opts.strip
        cmd = 'patch %s --output "%s"' % (popts, tmpfp.name)
        cin, cout = os.popen2(cmd, 'rw')
        cin.write('Index: %s\n' % filename)
        cin.write(patch)
        cin.close()

        # read output from patch.
        print(cout.read())
        if cout.close() != None:
            print("Error: running patch.", file=sys.stderr)

        #
        # spawn xxdiff
        #

        # compute stripped filename
        sfilename = filename
        if opts.strip:
            sfilename = os.sep.join(sfilename.split(os.sep)[opts.strip:])

        pno = 2
        leftfn, rightfn = sfilename, tmpfp.name

###FIXME not finished have to deal with deleted and added files

        # swap files if requested
        if opts.invert:
            leftfn, rightfn = rightfn, leftfn
            pno = 1

        # Invoke xxdiff.
        decision, mergedf, retcode = xxdiff.invoke.xxdiff_decision(
            opts, '--title%d' % pno, '%s (patched)' % sfilename,
            leftfn, rightfn)

        # print output of xxdiff command.
        if decision:
            print(decision)

        # perform the requested action
        if decision and not opts.dry_run:

            if decision == 'ACCEPT':
                if not opts.invert:
                    xxdiff.backup.backup_file(sfilename, opts)
                    shutil.copyfile(rightfn, sfilename)
                else:
                    assert rightfn == sfilename

            elif decision == 'REJECT':
                if not opts.invert:
                    xxdiff.backup.backup_file(sfilename, opts)
                    shutil.copyfile(leftfn, sfilename)
                else:
                    assert leftfn == sfilename

            elif decision == 'MERGED':
                xxdiff.backup.backup_file(sfilename, opts)
                shutil.copyfile(tmpfp.name, sfilename)

            elif decision == 'NODECISION':
                pass # do nothing

            else:
                raise SystemExit(
                        "Error: unexpected answer from xxdiff: %s" % decision)


def main():
    xxdiff.scripts.interruptible_main(patch_main)

if __name__ == '__main__':
    main()



# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
Functions for Perforce.
"""

__author__ = 'Martin Blais <blais@furius.ca>'


# stdlib imports.
import re
from subprocess import Popen, call, PIPE

# xxdiff imports.
from xxdiff.scripts import script_name


P4EXEC = 'g4' # Temporary.


def is_checked_out(filename):
    """
    Check if the file is already checked out.
    """
    p = Popen([P4EXEC, 'desc', '-n', filename], stdout=PIPE, stderr=PIPE, text=True)
    stdout, stderr = p.communicate()
    if p.returncode != 0:
        raise SystemExit( stderr )

    if re.search('opened for edit', stdout.strip()):
        return True

    return False

def checkout(filename):
    """
    Checkout the given filename for writable edit in Perforce.
    """
    call([P4EXEC, 'edit', filename])


def commit(filename):
    """
    Commit the given filename into a CL.
    """
    call([P4EXEC, 'change', filename])



# This file is part of the xxdiff package.  See xxdiff for license and details.
#
# Note: this has originally been distributed by Michalis Giannakidis, under the
# name cvsxxdiff.py.  This script was integrated in april 2006 in the xxdiff
# scripts.

"""xx-cvs-revcmp [<options>] [FILE ...]

Display file differences with the current or a previous CVS version.  Use one or
two of the version selection options to select which filesets to compare.


Here is the original submission email description:

  What, here in our company, find useful about cvsxxdiff is:

  - cvsxxdiff is one file. No python module directory or initialization is
    required.

  - Do a 'cvs up; cvsxxdiff' before commit to see all the local changes we have
    made.

  - Do 'cvs up; cvsxxdiff -b 1 file.c' to see the last chages that were made on
    a file by a colleague.

  - We often need to track bugs in our code. This means going back to previous
    revisions, thus doing ex: 'cvsxxdiff -b 1 -b 2 file.c' to compare the
    previous with the 2nd previous revision. (similar with cvsxxdiff -r 1.1 -r
    1.2 file.c)

  - The utility prints  in the standard output the logs between revisions i
    and j ( ex: cvsxxdif -b i -b j file.c)

  One script to do it all.

"""



__author__ = ('Michalis Giannakidis <mgiannakidis@gmail.com>',
              'Martin Blais <blais@furius.ca>')

# stdlib imports.
import sys, os, re, tempfile
from subprocess import Popen, PIPE, call
from os.path import *

# xxdiff imports.
import xxdiff.scripts
import xxdiff.invoke
from xxdiff.scripts import tmpprefix


def find_in_trunk(search_directories, fn):
    """
    Recursively search for paths of all files whose basename matches 'fn' under
    the directories from 'search_directories' and return a list of them.
    """
    # Make sure that fn is always just a basename.
    assert os.pathsep not in fn

    # Run the search.
    files = []
    for d in search_directories:
        for root, dirs, files_in_dir in os.walk(d):
            if 'CVS' in dirs:
                dirs.remove('CVS')  # don't visit CVS directories
            if fn in files_in_dir:
                files.append(join(root, fn))
    return files

match_status = re.compile('File:\s*([^\s]+)\s*Status: (.*)')

def collect_unupdated_files(diff_directories, resolve_conflicts):
    """
    Collect all files that are not updated (compared to the repository).
    """
    assert diff_directories

    p = Popen(['cvs', 'status'] + diff_directories, stdout=PIPE, text=True)
    stdout, stderr = p.communicate()
    lines = stdout.splitlines()

    collected_files = []
    for l in lines:
        m = match_status.search(l)
        if not m:
            continue

        if resolve_conflicts:
            if m.group(2) != 'File had conflicts on merge':
                continue
        else:
            if m.group(2) == 'Up-to-date':
                continue
        fn = m.group(1)

        # Run the search from the CWD if there are no diff dirs.
        files = find_in_trunk(diff_directories, fn)
        if files:
            if len(files) > 1:
                print('%s more than once' % fn)
                # FIXME: if we find a file more than once,
                # keep the unupdated.
            collected_files.extend(files)
        else:
            print('%s: No such file' % fn)

    return collected_files

def get_local_trunk_version(filename):
    """
    For filename, get the current revision number (of the local version).
    """
    dn, bn = split(filename)

    lines = open(join(dn, 'CVS', 'Entries')).readlines()
    match_unupdated = re.compile('/%s/([0-9\.]*)' % bn)
    for l in lines:
        m = match_unupdated.search(l)
        if  m:
            return m.group(1)
    return '1.1'

match_repository_rev = re.compile('Repository revision:\s([0-9\.]*)')

def get_repository_revision(filename):
    """
    Get the Repository revision of a file.
    """
    p = Popen(['cvs', 'status', filename], stdout=PIPE, text=True)
    stdout, stderr = p.communicate()
    for l in stdout.splitlines():
            m = match_repository_rev.search(l)
            if m:
                return m.group(1)
    return '1.1'

def get_previous_revision(filename, back):
    """
    For filename, get its previous revision number (compared to the local
    version).
    """
    r = get_local_trunk_version(filename).split('.')
    r[-1] = str(int(r[-1]) - back)
    r = '.'.join(r)
    return r

def get_revisions_between(r1, r2):
    """
    Return a list with the CVS version strings between to releases.
    """
    rev = []

    l_r1 = r1.split('.')
    l_r2 = r2.split('.')
    try:
        beg = int(l_r1[-1])
        end = int(l_r2[-1])
    except ValueError:
        return rev

    if end <= beg:
        rg = range(beg, end-1, -1)
    else:
        rg = range(beg, end+1)
    for r in rg:
        l_r1[-1] = str(r)
        rev.append('.'.join(l_r1))

    return rev

match_log_start = re.compile('^----------------------------')

def get_revision_log(filename, rev):
    """
    Get the CVS log of a filename given a revision.
    """
    log = []

    # Note: watch out for the 'cvs log' command, the value of the -r option must
    # not contain spaces before the value, it must be lumped together with the
    # option.  This is a CVS options parsing bug.
    p = Popen(['cvs', 'log', '-r%s' % rev, filename], stdout=PIPE, stderr=PIPE, text=True)
    stdout, stderr = p.communicate()
    # Note: we're simply throwing away stderr here.

    start_log = False
    for l in stdout.splitlines():
        if start_log == 0:
            if match_log_start.search(l):
                start_log = True
                continue
        if start_log:
            log.append(l)

    return log


def mkheader(filename):
    """
    Build a header for display.
    """
    return ('>%s<' % filename).center(80).replace(' ', '_')


def cvsxxdiff_bi_bj(diff_files, prevcounts):
    """
    Compare two relative revision numbers.
    """
    for fn in diff_files:
        print(mkheader(fn))

        # Get revision numbers.
        v1, v2 = [get_previous_revision(fn, prevcounts[x]) for x in [0, 1]]

        # Get the files from the server.
        tmpfiles = []
        for v in v1, v2:
            f = tempfile.NamedTemporaryFile(mode='w+', prefix=tmpprefix)
            tmpfiles.append(f)
            call(['cvs', 'update', '-r', v, '-p', fn], stdout=f)

        # Print the revisions.
        revs = get_revisions_between(v1, v2)
        for r in revs:
            print('\n'.join(get_revision_log(fn, r)))

        # Launch xxdiff.
        xxdiff.invoke.xxdiff_display(
            opts,
            '--title1', "%s ( %s )" % (fn, v1),
            '--title2', "%s ( %s )" % (fn, v2),
            *[x.name for x in tmpfiles])

def cvsxxdiff_bi(diff_files, prevcount):
    """
    Compare to one relative revision number.
    """
    for fn in diff_files:
        print(mkheader(fn))

        # Get revision numbers.
        v = get_previous_revision(fn, prevcount)
        vl = get_local_trunk_version(fn)

        # Print the revisions.
        revs = get_revisions_between(v, vl)
        for r in revs:
            print('\n'.join(get_revision_log(fn, r)))

        # Get the other file from the server.
        p = Popen(['cvs', 'update', '-r', v, '-p', fn], stdout=PIPE, text=True)

        # Launch xxdiff.
        xxdiff.invoke.xxdiff_display(
            opts,
            '--title1', ('%s ( %s ) File as %d versions ago' %
                         (fn, v, prevcount)),
            '--title2', ('%s ( %s ) local version' %
                         (fn, vl)),
            '-', fn, stdin=p.stdout)

def cvsxxdiff_ri(diff_files, action):
    """
    Compare to one absolute revision number.
    """
    for fn in diff_files:
        print(mkheader(fn))

        # Print the revisions.
        if action[0] == 'r':
            print('\n'.join(get_revision_log(fn, action[1])))

        # Launch xxdiff.
        p = Popen(['cvs', 'update', '-%s' % action[0], action[1], '-p', fn],
                  stdout=PIPE, text=True)

        xxdiff.invoke.xxdiff_display(
            opts,
            '--title1', '%s ( %s )'  % (fn, action[1]),
            '--title2', ('%s ( %s ) local version' %
                         (fn, get_local_trunk_version(fn))),
            '-', fn, stdin=p.stdout)

def cvsxxdiff_ri_rj(diff_files, actions):
    """
    Compare to two absolute revision numbers.
    """
    revisions = [x[1] for x in actions]
    for fn in diff_files:
        print(mkheader(fn))

        # Fetch the temporary files.
        tmpfiles = []
        for code, rev in actions:
            f = tempfile.NamedTemporaryFile(mode='w+', prefix=tmpprefix)
            tmpfiles.append(f)
            call(['cvs', 'update', '-%s' % code, rev, '-p', fn], stdout=f)

        # Print revisions.
        if actions[0][0] == 'r' and actions[1][0] == 'r':
            revs = get_revisions_between(revisions[0], revisions[1])
            for r in revs:
                print('\n'.join(get_revision_log(fn, r)))

        # Launch xxdiff.
        xxdiff.invoke.xxdiff_display(
            opts,
            '--title1', "%s ( %s )" % (fn, revisions[0]),
            '--title2', "%s ( %s )" % (fn, revisions[1]),
            *[x.name for x in tmpfiles])

def cvsxxdiff_rep(diff_files):
    """
    Compare checkout to repository.
    """
    for fn in diff_files:
        print(mkheader(fn))

        # Get the revisions.
        vr = get_repository_revision(fn)
        vl = get_local_trunk_version(fn)

        # Print the revisions.
        revs = get_revisions_between(vr, vl)
        for r in revs:
            print('\n'.join(get_revision_log(fn, r)))

        # Launch xxdiff.
        p = Popen(['cvs', 'update', '-p', fn], stdout=PIPE, text=True)

        xxdiff.invoke.xxdiff_display(
            opts,
            '--title1', '%s ( %s ) cvs version' % (fn, vr),
            '--title2', '%s ( %s ) local version' % (fn, vl),
            '-', fn, stdin=p.stdout)

def cvsxxdiff_c(diff_files):
    """
    Resolve conflicts.
    """
    for fn in diff_files:
        print(mkheader(fn))

        # Get revision numbers.
        vl = get_local_trunk_version(fn)
        print('\n'.join(get_revision_log(fn, vl)))


        # Launch xxdiff.
        xxdiff.invoke.xxdiff_display(
            opts,
            '--resource', 'Show.PaneMergedView: True',
            '--unmerge',
            fn)

def parse_options():
    """
    Parse the options.
    """
    import optparse
    parser = optparse.OptionParser(__doc__.strip())

    group = optparse.OptionGroup(parser, "Options for selecting the revisions",
                                 "These options allow you to specify the "
                                 "revisions to compare.  You can specify zero, "
                                 "one or two of the following options.  The "
                                 "default is otherwise HEAD (latest).")

    group.add_option('-r', '--revision-tag', '--tag', action='append',
                     metavar='TAG', dest='revisions', default=[],
                     help="A tag like CVS.  See cvs(1).")

    group.add_option('-D', '--date-spec', action='append',
                     dest='dates', default=[],
                     help="A date spec like CVS.  See cvs(1).")

    group.add_option('-b', '--nb-revisions-before', action='append', type='int',
                     metavar='NBREVS', dest='before', default=[],
                     help="The number of previous CVS revisions from BASE.")

    parser.add_option_group(group)

    parser.add_option('-c', '--resolve-conflicts', action='store_true',
                      help="Resolve conflicts in merged files.")

    xxdiff.invoke.options_graft(parser)

    xxdiff.scripts.install_autocomplete(parser)
    opts, args = parser.parse_args()

    xxdiff.invoke.options_validate(opts, parser)

    # Figure out the two actions from the options that were passed in.  There
    # can be at most two actions.  Each action is specified as a pair of
    # (action-code, spec).
    actions = []
    for code, optlist in (('b', opts.before),
                          ('r', opts.revisions),
                          ('D', opts.dates)):
        actions.extend((code, x) for x in optlist)

    if len(actions) > 2:
        parser.error("You cannot specify more than two revision selection "
                     "options.")

    return opts, actions, args


def revcomp_main():
    """
    Command-line interface (looks at sys.argv to decide what to do).
    """
    global opts
    opts, actions, filenames = parse_options()

    # Separate files and directories in arguments.
    diff_files, diff_directories = [], []
    for fn in filenames:
        if isfile(fn):
            diff_files.append(fn)
        elif isdir(fn):
            diff_directories.append(fn)

    # Collect files.
    #
    # If no files were given on the command line or command was invoked on
    # specific directories, search for unupdated files recursively in
    # o) in the specified directories or
    # o) the current dir
    if len(diff_files) == 0 or len(diff_directories) :
        collected_files = collect_unupdated_files(diff_directories or ['.'],
                                                  opts.resolve_conflicts)
        diff_files.extend(collected_files)

    if len(diff_files) > 1 :
        # Print a message to the user about the list of files to diff.
        # Print it in a single line. This allows to copy paste the output to
        # ex. a 'cvs commit' command.
        print('%d files to diff:' % len(diff_files))
        print('=' * 80)
        print(' '.join(diff_files))
        print('=' * 80)

    # Dispatch to the appropriate method.
    if opts.resolve_conflicts:
        # ex: -c
        if actions:
            raise SystemExit("You cannot specify revision selectors when "
                             "resolving conflicts.")
        cvsxxdiff_c(diff_files)

    # With a single revision selector.
    elif len(actions) == 1:
        code, count = actions[0]
        if code == 'b':
            # ex: -b1
            cvsxxdiff_bi(diff_files, count)

        elif code in ('r', 'D'):
            # ex: -r1.2  # or with date_spec
            cvsxxdiff_ri(diff_files, actions[0])

        else:
            raise SystemExit("Unsupported revision combination.")

    # With two revision selectors.
    elif len(actions) == 2:
        codes = tuple([x[0] for x in actions])
        revs = tuple([x[1] for x in actions])

        if codes == ('b', 'b'):
            # ex: -b1 -b2
            cvsxxdiff_bi_bj(diff_files, revs)

        if codes[0] in ('r', 'D') and codes[1] in ('r', 'D'):
            # ex: -r1.2 -r1.3
            cvsxxdiff_ri_rj(diff_files, actions)

        else:
            raise SystemExit("Unsupported revision combination.")

    elif not actions:
        # ex: cvsxxdiff
        cvsxxdiff_rep(diff_files)


def main():
    xxdiff.scripts.interruptible_main(revcomp_main)

if __name__ == '__main__':
    main()



# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
Functions to invoke xxdiff in various ways.
"""



__author__ = 'Martin Blais <blais@furius.ca>'

# stdlib imports.
import os, optparse, tempfile, logging, io
from os.path import *
from subprocess import Popen, PIPE

# xxdiff imports.
from xxdiff.scripts import tmpprefix


def options_graft(parser):
    """
    Graft options on given parser for invocations of xxdiff.
    """
    group = optparse.OptionGroup(parser, "Options for xxdiff subprocesses",
                                 "These options affect how xxdiff is called.")

    group.add_option('-Z', '--xxdiff-exec', action='store', metavar='XXDIFF',
                     default='xxdiff',
                     help="Specifies which xxdiff executable to use (default "
                     "is looked up in PATH.")

    group.add_option('-O', '--xxdiff-option', '--xxdiff-opt',
                     action='append', metavar="OPTS", dest='xxdiff_options',
                     default=[],
                     help="Additional option to pass on to xxdiff.  You "
                     "can invoke this many times.")

    group.add_option('-Y', '--xxdiff-verbose', action='store_true',
                     help="Output xxdiff commands on stdout, for debugging.")

    parser.add_option_group(group)

    return group


def options_validate(opts, parser, logs=None):
    """
    Validate xxdiff options.
    """
    pass


def _run_xxdiff(cmd, opts, stdin):
    """
    Runs the given xxdiff command and return a Popen object.
    'stdin' is the optional stdin text or open file to send to xxdiff.

    Note: this is an internal function, not meant to be called from the outside.
    """
    if getattr(opts, 'xxdiff_verbose', None):
        print('===', ' '.join(cmd))

    if stdin is not None:
        assert '-' in cmd
    if isinstance(stdin, str):
        # stdin is text.
        intype = PIPE
        intext = stdin
    elif isinstance(stdin, io.IOBase) or hasattr(stdin, 'read'):
        # stdin is an open pipe/file.
        intype = stdin
    else:
        intype = None

    try:
        p = Popen(cmd,
                  stdout=PIPE,
                  stderr=PIPE,
                  stdin=intype,
                  text=True)
    except OSError as e:
        extramsg = ''
        if e.errno == 2:
            extramsg = '\nHint: Check if xxdiff is accessible in your path.'
        raise SystemExit("Error: running xxdiff '%s'" % e + extramsg)

    # Write the given text to stdin if necessary.
    if intype is PIPE:
        p.stdin.write(intext)

    return p

decisions = ('ACCEPT', 'REJECT', 'MERGED', 'NODECISION')

def xxdiff_decision(opts, *arguments, **kwds):
    """
    Runs xxdiff with the given arguments, passed directly to subprocess.call(),
    and return the decision code, the temporary file containing merged output,
    and xxdiff's return code.  We create this temporary file to be used as the
    merged output to be collected and modify the options accordingly.  It is a
    named file.  If you close the merged temporary file, it is deleted
    automatically.  This file is opened in 'rw' mode, so you can rewind it and
    read it directly if desired, or you can use its filename with high-level
    operations, like file copy, as long as you keep the reference to the file.

    You should unpack its arguments like this::

       decision, mergedf, retcode = xxdiff_decision( ... )

       # use mergedf.name to copy the file or seek(0) and read() it again.

    If the decision code is 'NODECISION', the merged file is set to None (its
    contents are undefined).

    'stdin' (in kwds) can be used to pass text or an open file to the
    subordinate xxdiff process.

    If you want to perform actions asynchronously while xxdiff is waiting for
    the user's decision, you can set 'nowait' (in kwds) to True, the function
    then returns immediately after starting xxdiff, and a proxy callable object
    is returned that can be invoked to wait for the final results and returns
    the same data as the function specifies.  You use it like this::

      xxwait = xxdiff_decision( ..., nowait=1)

      # Do something...

      # Wait for the xxdiff results
      decision, merged, retcode = xxwait()

    """
    # Create a temporary file to contain the output or merged results.
    mergedf = tempfile.NamedTemporaryFile(mode='w+', prefix=tmpprefix)

    # Get the appropriate xxdiff executable and options.
    xexec = getattr(opts, 'xxdiff_exec', 'xxdiff')
    options = getattr(opts, 'xxdiff_options', [])

    # Make sure that xxdiff is invoked with the decision switch.
    if '--decision' not in options:
        options.insert(0, '--decision')

    assert '--merged-filename' not in (list(arguments) + options), (list(arguments) + options)
    alloptions = options + ['--merged-filename', mergedf.name] + list(arguments)

    # If we're not waiting, we only want to return after all the input has been
    # processed.  This is always the case.
    nowait = kwds.pop('nowait', None)
    if nowait and '--indicate-input-processed' not in alloptions:
        alloptions.insert(0, '--indicate-input-processed')

    # Run xxdiff.
    cmd = [xexec] + alloptions
    p = _run_xxdiff(cmd, opts, kwds.pop('stdin', None))

    # Define waiter object.
    def waiter():
        # Select-wait for stdout and stderr
        stdout, stderr = p.communicate()

        # If xxdiff failed, we bail out of the script.
        if p.returncode == 2:
            logging.error("Error: running xxdiff as '%s'.\n" %
                          ' '.join(cmd) + stderr)
            return 'NODECISION', None, -1

        # Get the decision code from xxdiff.
        lines = stdout.splitlines()
        if not lines:
            logging.error("Error: running xxdiff as '%s'.\n" %
                          ' '.join(cmd) + stderr)
            return 'NODECISION', None, -1

        decision = lines[0].strip()
        assert decision in decisions

        if decision == 'NODECISION':
            mergedf.close()
            out_mergedf = None
        else:
            out_mergedf = mergedf

        return decision, out_mergedf, p.returncode

    # Eat up the processed tag.
    #
    # Important note: there is a subtle race condition that can cause an error
    # here, but it is rather unlikely to happen (in practice it will never
    # happen because xxdiff is faster than the user's fingers).
    if nowait and p.returncode is None:
        # Wait for the processed tag.
        line = p.stdout.readline().strip()
        if line != 'INPUT-PROCESSED':
            stdout, stderr = p.communicate()
            logging.error("Error: running xxdiff as '%s'.\n" %
                          ' '.join(cmd) + stderr)

        return waiter

    # Wait for the results and return them.
    return waiter()


def xxdiff_display(opts, *arguments, **kwds):
    """
    Runs xxdiff with the given arguments, passed directly to subprocess.call().
    We do not run it in decision mode.  We return xxdiff's own return code (the
    --exit-with-merge-status will have an effect on this).  There is a similar
    'nowait' option as xxdiff_decision() with a waiter object returned to wait
    on the child process.  Otherwise, this function will wait for the child to
    terminate.

    'stdin' (in kwds) can be used to pass text or an open file to the
    subordinate xxdiff process.  You will need to specify a '-' as one of the
    file arguments to specify on which side to display the stdin file.

    We do not create a merged file, but you are free to pass in these options if
    so desired.
    """
    if '--decision' in arguments:
        raise RuntimeError("Internal error: use xxdiff_decision() "
                           "instead of xxdiff_display() if you want a user "
                           "return code.")

    # Get the appropriate xxdiff executable and options.
    xexec = getattr(opts, 'xxdiff_exec', 'xxdiff')
    options = getattr(opts, 'xxdiff_options', [])

    alloptions = options + list(arguments)

    # If we're not waiting, we only want to return after all the input has been
    # processed.  This is always the case.
    nowait = kwds.pop('nowait', None)
    if nowait and '--indicate-input-processed' not in alloptions:
        alloptions.insert(0, '--indicate-input-processed')

    # Run xxdiff.
    cmd = [xexec] + alloptions
    p = _run_xxdiff(cmd, opts, kwds.pop('stdin', None))

    # Define waiter object.
    def waiter():
        # Select-wait for stdout and stderr
        stdout, stderr = p.communicate()

        # If xxdiff failed, we bail out of the script.
        if p.returncode == 2:
            logging.error("Error: running xxdiff as '%s'.\n" %
                          ' '.join(cmd) + stderr)
            return -1

        return p.returncode

    # Eat up the processed tag.
    #
    # Important note: there is a subtle race condition that can cause an error
    # here, but it is rather unlikely to happen (in practice it will never
    # happen because xxdiff is faster than the user's fingers).
    if nowait and p.returncode is None:
        # Wait for the processed tag.
        line = p.stdout.readline().strip()
        if line != 'INPUT-PROCESSED':
            stdout, stderr = p.communicate()
            logging.error("Error: running xxdiff as '%s'.\n" %
                          ' '.join(cmd) + stderr)

        return waiter

    # Wait for the results and return them.
    return waiter()


def title_opts(*titles):
    """
    Generate title options for each of the given titles.  This returns a list of
    the options to give xxdiff to set the titles in the given order.
    """
    assert len(titles) <= 3
    topts = []
    for idx, title in enumerate(titles):
        topts.append('--title%d' % (idx+1))
        topts.append(title)
    return topts


def test():
    """
    Test launcher.
    """
    import optparse
    parser = optparse.OptionParser()
    opts, args = parser.parse_args()

    class Opts:
        xxdiff = 'xxdiff'

    f1, f2, f3 = [join('/home/blais/p/xxdiff/test', x) for x in ('mine', 'older', 'yours')]

    for t in args:
        if t == 'simple':
            print(xxdiff_decision(Opts, f1, f2))

        elif t == 'async':
            w = xxdiff_decision(Opts, f1, f2, nowait=1)
            print('Deleting temp input files...')
            print(w())

        elif t == 'error':
            print(xxdiff_decision(Opts, '--bullshit', f1, f2))

        elif t == 'error-nowait':
            print(xxdiff_decision(Opts, '--bullshit', f1, f2, nowait=1))

        if t == 'display':
            print(xxdiff_display(Opts, f1, f2))

        if t == 'display-nowait':
            w = xxdiff_display(Opts, f1, f2, nowait=1)
            print('Deleting temp input files...')
            print(w())

        if t == 'pipe-text':
            print(xxdiff_display(Opts, f2, '-', stdin='Some text\n'))

        if t == 'pipe-cmd':
            p = Popen('cat $HOME/.xxdiffrc', shell=True, stdout=PIPE, text=True)
            print(xxdiff_display(Opts, '-', f2, stdin=p.stdout))

if __name__ == '__main__':
    test()



# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
Common command-line/script -related code for all the scripts.

This directory contains the actual scripts that are run from the stubs.  Any
significant code should be shared in the parent directory.  These files should
contain nothing but the high-level algorithms and the options parsing for the
scripts.
"""

__author__ = 'Martin Blais <blais@furius.ca>'


# stdlib imports.
import sys
from os.path import basename


# Check interpreter version requirements.
if sys.version_info[:2] < (2, 4):
    raise SystemExit(
        "Error: you need at least version 2.4 of Python or greater.")


def interruptible_main(mainfun):
    """
    Wrapper to allow keyboard interrupt signals.
    """
    try:
        mainfun()
    except KeyboardInterrupt:
        # If interrupted, exit nicely
        raise SystemExit('Interrupted.')


def install_autocomplete(parser):
    """
    Install automatic programmable completion support if available.
    Scripts should work even if it's not there.
    """
    try:
        import optcomplete
        optcomplete.autocomplete(parser)
    except ImportError:
        pass


def passthruopts(argv):
    """
    Hack to allow passing some options.  This hack may fail when the given
    options are incorrect but has the advantage that it doesn't require
    knowledge of a specific xxdiff command-line (it will not rot as fast).
    """
    try:
        i = argv.index('--endopts')
        opts, args = argv[1:i], argv[i+1:]
    except ValueError:
        for i in range(len(argv)-1, -1, -1):
            if argv[i].startswith('-'):
                break
        opts, args = argv[1:i+1], argv[i+1:]
    return opts, args


# Default prefix for temporary files uses the invoking script's name.
script_name = basename(sys.argv[0])
tmpprefix = '%s.' % script_name




# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
Functions for spawning editor windows.
"""

__author__ = "Martin Blais <blais@furius.ca>"


# stdlib imports.
import os, tempfile
from os.path import *
from subprocess import Popen, PIPE

# xxdiff imports.
from xxdiff.scripts import tmpprefix


__all__ = ('spawn_editor',)


def_editor = ["xterm", "-e", '/usr/bin/vi "%s"']

def spawn_editor(initcontents=None, filename=None):
    """
    Spawns an editor window and returns a waitable object that will block until
    the editor is done and recuperate the text once it's done.  If you delete
    the returned object without waiting on it, the editor program is killed.
    
    If 'initcontents' is given, it is inserted in the temporary file to be
    edited before spawing the editor.  This is only used if no 'filename' has
    been specified.

    If 'filename' is specified, the editor is spawned on that file rather than a
    temporary.  If the file specified by filename exists and is not empty, we
    recycle it.

    This function returns the contents of the edited file, or None, if the edit
    was cancelled.  It may return an empty string.
    """
    # Create and the filename that we will eventually read from.
    tmpf = None
    if filename is not None:
        # Check if the specified file already exists, and don't overwrite it if
        # this is the case.
        if exists(filename) and getsize(filename) > 0:
            tmpf = open(filename, 'r')
            initcontents = None
        else:
            tmpf = open(filename, 'w')
    else:
        tmpf = tempfile.NamedTemporaryFile(mode='w', prefix=tmpprefix)
        filename = tmpf.name
        
    # Initialize the contents of the file if requested.
    if initcontents is not None:
        assert isinstance(initcontents, str)
        tmpf.write(initcontents)
        tmpf.flush()

    # Compute the command to spawn to launch the editor.
    #
    # Note: the editor should the kind to open a new window because we're going
    # to keep printing stuff to stdout during the diffing.  You can just set it
    # up to spawn a new VT if you need to, like in an xterm or something.
    for var in 'XXDIFF_EDITOR', 'SVN_EDITOR', 'EDITOR':
        editor = os.environ.get(var, None)
        if editor:
            break

    shell = False
    if editor:
        if '%s' in editor:
            editor %= filename
            cmd = editor
            shell = True
        else:
            cmd = [editor, filename]
    else:
        cmd = def_editor
        cmd[-1] %= filename

    p = Popen(cmd, shell=shell, stdout=PIPE, stderr=PIPE, text=True)
    
    def waiter():
        "Waiter closure."
        stdout, stderr = p.communicate()
        if stderr:
            raise RuntimeError("Error: running editor:\n%s\n" % stderr)
        writtenf = open(tmpf.name, 'r')
        text = writtenf.read()
        writtenf.close()
        tmpf.close()
        return text

    # Set the command string for the client to display if necessary.
    waiter.command = cmd
    if isinstance(waiter.command, (tuple, list)):
        waiter.command = ' '.join(waiter.command)

    return waiter




# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
xx-pyline [<options>] <python-expr> [<python-expr> ...]

Run the given Python code on each line of the input.  This is a bit like sed,
except that it's just Python code.

  Inspired by Graham Fawcett's Python Cookbook recipe:
  http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/437932

    'This utility was born from the fact that I keep forgetting how to use
    sed, and I suck at Perl. It brings ad-hoc command-line piping
    sensibilities to the Python interpeter.'

This program walks a directory hierarchy, selects some files to be processed by
filenames (or by default, process all files found), then the given action is
performed on the selected file, compared, if some changes were carried out, a
graphical diff is presented to ask the user whether he wants to apply the
changes or not.
"""

__author__ = "Martin Blais <blais@furius.ca>"
__depends__ = ['xxdiff', 'Python-2.4']


# stdlib imports.
import sys, os, traceback
from os.path import *

# xxdiff imports.
import xxdiff.scripts
import xxdiff.xformloop


class PylineTransformer(xxdiff.xformloop.Transformer):
    """
    Transformer that greps the file for a regular expression, and if it matches,
    runs it through a sed command.
    """
    def __init__(self, opts, expr):
        xxdiff.xformloop.Transformer.__init__(self, opts)

        # Try to compile the given expression.
        try:
            self.codeobj = compile(expr, 'command', 'eval')
        except Exception:
            raise SystemExit("Error: could not compile expression:\n%s" %
                             traceback.format_exc(limit=0))

    def transform(self, fn, outf):
        # Open and read input file in memory.
        try:
            inf = open(fn, 'r')
        except IOError as e:
            raise SystemExit("Error: Could not read file '%s':\n  %s" % (fn, e))

        # Run the code on each line.
        modified = False
        for numz, line in enumerate(inf):
            line = line[:-1]
            num = numz + 1
            words = [w for w in line.strip().split(' ') if len(w)]
            try:
                result = eval(self.codeobj, globals(), locals())
            except Exception:
                raise SystemExit("Error: running expression on line %d:\n%s\n%s"
                                 % (numz, line, traceback.format_exc(limit=0)))

            if result is None or result is False:
                modified = True
                continue
            elif isinstance(result, list) or isinstance(result, tuple):
                result = ' '.join(map(str, result))
            else:
                result = str(result)

            if not modified and line != result:
                modified = True

            outf.write(result)
            if not result.endswith('\n'):
                outf.write('\n')

        return modified


def parse_options():
    import optparse
    parser = optparse.OptionParser(__doc__.strip())

    parser.add_option('-m', '--import-lib', action='append',
                      default=[],
                      help="Import the given module before running the script.")

    opts, args, selector = xxdiff.xformloop.parse_args(parser)

    # Check that we got two arguments
    expr = ' '.join(args)
    if not expr.strip():
        expr = 'line'  # no-op

    # Force to always perform a diff on output.
    opts.verbose = 2

    # Import the necessary libraries.
    for libname in opts.import_lib:
        try:
            globals()[libname] = __import__(libname.strip())
        except ImportError as e:
            parser.error("Importing module '%s': %s" % (libname, e))

    return expr, opts, selector


def pyline_main():
    """
    Main program for pyline script.
    """
    # Parse the options.
    expr, opts, selector = parse_options()

    # Create an appropriate transformer.
    xformer = PylineTransformer(opts, expr)

    try:
        decision_map = xxdiff.xformloop.transform_replace_loop(
            opts, selector, xformer, sys.stdout)
    finally:
        xxdiff.xformloop.postloop_footer(opts)

def main():
    xxdiff.scripts.interruptible_main(pyline_main)

if __name__ == '__main__':
    main()



# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
Functions to checkout files from various SCM systems before writing to them.
ClearCase requires that, and there is a configuration option in CVS that
requires 'cvs edit' too.  This is where the director code goes.  Code for
specific SCM systems is located under in the xxdiff.scm package.
"""

__author__ = 'Martin Blais <blais@furius.ca>'


# stdlib imports.
import optparse


checkout_choices = ('clearcase', 'subversion', 'cvs', 'perforce')

def options_graft(parser):
    """
    Graft options on given parser for checkout options.
    """
    group = optparse.OptionGroup(parser, "SCM options",
                                 "These options are related to source code "
                                 "management systems you may be using.")

    group.add_option('-C', '--checkout', action='store',
                     type='choice', choices=checkout_choices,
                     metavar="NAME",
                     help="Checkout files from an SCM before overwriting. "
                     "Choices include: %s" % ', '.join(checkout_choices))

    parser.add_option_group(group)

    return group

def options_validate(opts, parser, logs=None):
    """
    Validate checkout options.
    """
    pass # Nothing to do.


def get_module(modname):
    """
    Returns the appropriate module for the given SCM.
    """
    try:
        mod = __import__("xxdiff.scm.%s" % modname)
        mod = getattr(mod.scm, modname)
    except ImportError as e:
        raise SystemExit("Internal Error: SCM module not found.")
    return mod

def insure_checkout(fn, opts, logs=None):
    """
    Check the file out from the SCM, if necessary.
    Returns 'True' if the file was checked out.
    """
    scm_module = get_module(opts.checkout)

    # Verify if the file is already checked out.
    if scm_module.is_checked_out(fn):
        return False

    # Checkout the file.
    if logs:
        logs.write("Checking out file '%s'" % fn)

    scm_module.checkout(fn)
    return True



# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
xx-filter [<options>] <cmd>

Run a command to perform replacements in files and confirm the application of
changes via xxdiff.

This program walks a directory hierarchy, selects some files to be processed by
filenames (or by default, process all files found), then the given filter
command is run with the selected file as input and its output captured,
compared, if the files differ, a graphical diff is presented to ask the user
whether he wants to apply the changes or not.
"""

__author__ = "Martin Blais <blais@furius.ca>"
__depends__ = ['xxdiff', 'Python-2.4']


# stdlib imports.
import sys, os, filecmp
from os.path import *
from subprocess import Popen, PIPE

# xxdiff imports.
import xxdiff.scripts
import xxdiff.xformloop


class FilterTransformer(xxdiff.xformloop.Transformer):
    """
    Transformer that greps the file for a regular expression, and if it matches,
    runs it through a sed command.
    """
    def __init__(self, opts, filtcmd):
        xxdiff.xformloop.Transformer.__init__(self, opts)
        self.filtcmd = filtcmd

    def transform(self, fn, outf):
        # Open input file.
        inf = open(fn, 'r')

        # Run the user's command.
        p = Popen(self.filtcmd, shell=True, stdin=inf, stdout=outf, stderr=PIPE, text=True)

        # Note: we do not check the return code, because we do not know the
        # client's convention for the program he invokes.
        returncode = p.wait()

        # Hmmm... what should we do with errors?
        errors = p.stderr.read()

        # Check if the files differ and indicate to process if so.
        #
        # Note: using filecmp should be faster than intercepting and reading all
        # the files in memory, because hopefully they are only reading and
        # comparing chunks at a time to bail out when there is a difference
        # instead of reading the whole data in memory before comparison.
        return not filecmp.cmp(fn, outf.name)


def parse_options():
    import optparse
    parser = optparse.OptionParser(__doc__.strip())

    opts, args, selector = xxdiff.xformloop.parse_args(parser)

    # Check that we got two arguments
    if len(args) != 1:
        parser.error("You must specify a filter command to run on the files.")
    filtcmd = args[0]

    # Force to always perform a diff on output.
    opts.verbose = 2

    return filtcmd, opts, selector


def filter_main():
    """
    Main program for find-grep-sed script.
    """
    # Parse the options.
    filtcmd, opts, selector = parse_options()

    # Create an appropriate transformer.
    xformer = FilterTransformer(opts, filtcmd)

    try:
        decision_map = xxdiff.xformloop.transform_replace_loop(
            opts, selector, xformer, sys.stdout)
    finally:
        xxdiff.xformloop.postloop_footer(opts)


def main():
    xxdiff.scripts.interruptible_main(filter_main)

if __name__ == '__main__':
    main()




# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
Functions that contain stuff related to PostgreSQL.

In particular, functions to dump and parse a PostgreSQL database schema.  We use
this to perform schema comparison in one of the scripts.
"""



__author__ = "Martin Blais <blais@furius.ca>"


# stdlib imports.
import os, optparse, re
from os.path import *
from subprocess import Popen, PIPE

# xxdiff imports.
from xxdiff.utils import consepairs


def options_graft(parser):
    """
    Graft options on given parser for SQL connections.
    """
    group = optparse.OptionGroup(parser, "SQL connection options",
                                 "These options affect how we connect to "
                                 "SQL databases.")

    group.add_option('-H', '--host', action='store',
                     default=None,
                     help="Database server host or socket directory")

    group.add_option('-p', '--port', action='store',
                     default=None,
                     help="Database server port number")

    group.add_option('-U', '--username', '--user', action='store',
                     default=None,
                     help="Connect as specified database user")

    parser.add_option_group(group)

    return group


def options_validate(opts, parser, logs=None):
    """
    Validate replacement options.
    """
    args = []
    if opts.host:
        args.extend( ['-h', opts.host] )
    if opts.port:
        args.extend( ['-p', opts.port] )

    opts.pgsqlargs = args


dbspec_re = re.compile(
    '^(?:([a-zA-Z0-9_]+)@)?([a-zA-Z0-9_]+)(?:/([a-zA-Z0-9_]+))?$')

def parse_dbspec(dbspec, parser, opts):
    """
    Parse the DB spec and return the user, dbname, schema components.

    DB specs are of the form:  [user@]<dbname>[/schema]
    """
    mo = dbspec_re.match(dbspec)
    if not mo:
        parser.error("Invalid database/schema specification: '%s'" % dbspec)

    user, db, schema = mo.groups()

    # Set defaults from the options.
    if user is None:
        user = opts.username

    return Schema(dbspec, user, db, schema)


class Schema(object):
    """
    Container for all database schema-related things.
    """
    def __init__(self, dbspec, user, dbname, schema):
        self.dbspec = dbspec
        self.user, self.dbname, self.schema = user, dbname, schema
        self.dump = None

def dump_schema(user, dbname, schema, opts):
    """
    Dump the schema for the given database.  Returns the dump text on output.
    """
    tmplcmd = ['pg_dump', '--schema-only'] + opts.pgsqlargs

    cmd = list(tmplcmd)
    if user:
        cmd.extend(['-u', user])
    if schema:
        cmd.extend(['--schema', schema])
    cmd.append(dbname)

    p = Popen(cmd, stdout=PIPE, text=True)
    dump, stderr = p.communicate()
    if p.returncode != 0:
        raise SystemExit("Error: Dumping database '%s':\n%s\n" %
                         (dbname, stderr))

    return dump


sec_re = re.compile('^-- (?:Data for )?Name:\s*([^\s;]+);\s*Type:\s*([^;]+);(.*)$', re.M)
com_re = re.compile('--.*$', re.M)
ct_re = re.compile('^CREATE TABLE.*?(\\().*?(\\);)', re.M|re.S)

def parse_dump(dbdump, sort_columns=False):
    """
    Parse a PostgreSQL database dump, extracting all its section into a list of
    (name, type, contents) tuples. The entries are built from the Name and Type
    fields in the header comments. Contents of data chunks for tables are
    returned as type 'DATA', so if you're going to turn this into a map, be
    careful to filter out the 'data' entries or concatenate them to the schema
    (whichever is appropriate for your application).
    """
    # Class to contain info about chunks.
    class Chunk:
        def __init__(self, mo):
            self.mo = mo
            self.name, self.typ = mo.group(1, 2)

    # Parse chunks.
    chunks = [Chunk(mo) for mo in sec_re.finditer(dbdump)]
    for c1, c2 in consepairs(chunks):
        c1.contents = dbdump[c1.mo.end():c2.mo.start()]
    chunks[-1].contents = dbdump[chunks[-1].mo.end():]

    # Remove comments in the contents.
    for c in chunks:
        c.descline = c.mo.group(0)
        c.contents = com_re.sub('', c.contents).strip() + '\n'
        c.mo = None # release the match objects

        # Sort columns in CREATE TABLE statements.
        if sort_columns:
            mo = ct_re.match(c.contents)
            if mo:
                pre, post = c.contents[:mo.end(1)], c.contents[mo.start(2):]
                columns = c.contents[mo.end(1):mo.start(2)].strip()
                line_cols = [x.endswith(',') and x or '%s,' % x for x in list(map(str.strip, columns.splitlines()))]
                line_cols.sort()
                c.contents = (pre + '\n' +
                              ''.join('   %s\n' % x for x in line_cols) +
                              post)

    return list((c.name, c.typ, c.descline + '\n\n' + c.contents)
                for c in chunks)

def dump_chunks(chunks):
    """
    Debug printout of chunks.
    """
    for chunk in chunks:
        print('========', chunk.name, chunk.typ)
        print(chunk.descline)
        print(chunk.contents)



# This file is part of the xxdiff package.  See xxdiff for license and details.

"""
xx-rename [<options>] <from-str> <to-str> [<from-str> <to-str> ...]

Run a command to perform a replacement of <from-str> to <to-str> in files
and confirm the application of changes via xxdiff.

This program walks a directory hierarchy, selects some files to be processed by
filenames (or by default, process all files found), then the given renaming is
performed on the selected file, compared, if some renaming was applied, a
graphical diff is presented to ask the user whether he wants to apply the
changes or not.

Note that if you specify multiple from/to pairs as arguments, all the renamings
will be performed simultaneously, in the order that they are given.
"""

__author__ = "Martin Blais <blais@furius.ca>"
__depends__ = ['xxdiff', 'Python-2.4']


# stdlib imports.
import sys, os, re
import logging
from os.path import *

# xxdiff imports.
import xxdiff.scripts
import xxdiff.xformloop
from xxdiff.utils import idify


class RenameTransformer(xxdiff.xformloop.Transformer):
    """
    Transformer that greps the file for a regular expression, and if it matches,
    runs it through a sed command.
    """
    def __init__(self, opts, renames):
        """
        'renames' should be a list of (from, to) pairs.
        """
        xxdiff.xformloop.Transformer.__init__(self, opts)

        # Compile the from strings as regexps.
        self.renames = []
        for sfrom, sto in renames:

            # Escape the string for regexp compilation if necessary.
            if self.opts.regexp:
                regexp = sfrom
            else:
                # Escape the special chars
                regexp = re.escape(sfrom)

            # Compile the regular expression.
            try:
                refrom = re.compile(regexp, re.MULTILINE)
            except re.error as e:
                raise SystemExit(
                    "Error: Could not compile given regexp '%s':\n%s" %
                    (regexp, e))

            self.renames.append( (refrom, sto) )

    def transform(self, fn, outf):
        # Open and read input file in memory.
        try:
            inf = open(fn, 'r')
            text = inf.read()
        except (IOError, UnicodeDecodeError) as e:
            logging.info("Error: Could not read file '%s':\n  %s", (fn, e))
            return False

        # Replace the string or regexp.
        nbrepl = 0
        for refrom, sto in self.renames:
            text, n = refrom.subn(sto, text)
            nbrepl += n

        # If there were no replacements, skip the file.
        if nbrepl == 0:
            return False

        outf.write(text)
        return True


def parse_options():
    import optparse
    parser = optparse.OptionParser(__doc__.strip())

    parser.add_option('-R', '--regexp', '--re', action='store_true',
                      help="Interpret <from-string> as a regular expression.")

    ## FIXME: TODO, implement this one day, this would be tremendously cool.
    ##     parser.add_option('-e', '--emacs-case', '--preserve-case',
    ##                       action='store_true',
    ##                       help="Perform smart replacements with respect to "
    ##                       "case-sensitivity like emacs does.")

    opts, args, selector = xxdiff.xformloop.parse_args(parser)

    # Check that we got 2, 4, 6, 8, ... arguments
    if len(args) < 2 or len(args) % 2 != 0:
        parser.error("You must specify an even number of <from-string> and "
                     "<to-string> pairs to rename.")
    renames = []
    it = iter(args)
    for sfrom in it:
        renames.append( (sfrom, next(it)) )

    # Force to always perform a diff on output.
    opts.verbose = 2

    # Idify the arguments from the first rename pair and setup a backup suffix,
    # to make the default backup directory easier to identify.
    sfrom, sto = renames[0]

    if opts.backup_prefix is None:
        # Set a better default for the backup prefix.
        opts.backup_prefix = '%s.%s' % tuple(map(idify, (sfrom, sto)))

    return renames, opts, selector


def rename_main():
    """
    Main program for renaming script.
    """
    # Parse the options.
    renames, opts, selector = parse_options()

    # Create an appropriate transformer.
    xformer = RenameTransformer(opts, renames)

    try:
        decision_map = xxdiff.xformloop.transform_replace_loop(
            opts, selector, xformer, sys.stdout)
    finally:
        xxdiff.xformloop.postloop_footer(opts)

def main():
    xxdiff.scripts.interruptible_main(rename_main)

if __name__ == '__main__':
    main()
